<h1 class="w-100 text-center" style="padding-top: 10%;">Coming Soon</h1>
        <p class="w-100 text-center fs-6">Sign up for our newsletter and we will inform you when the next new functionalities are released.<br><br>
        <a href="https://slurpit.io/getting-started/#subscribe" target="_blank" class="btn btn-primary">Subscribe</a></p>