# Generated by Django 4.2.8 on 2024-01-26 21:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='slurpitplanning',
            name='name',
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='slurpitplanning',
            name='planning_id',
            field=models.BigIntegerField(unique=True),
        ),
    ]
