{% load buttons %}
{% load i18n %}
<div class="show {% if request.GET.tab == 'onboarded' %}active{% endif %}" id="onboarded-device-list" role="tabpanel" aria-labelledby="onboarded-device-list-tab">
  {# Object table controls #}
  {% include 'inc/table_controls_htmx.html' with table_modal="ObjectTable_config" %}
  <form method="post" class="form form-horizontal">
    {% csrf_token %}
    {# Object table controls #}
    {# Form buttons #}
    <div class="noprint bulk-buttons">
      <div class="bulk-button-group">
          {% block bulk_buttons %}
          <div class="" role="group">
              <button type="submit" name="_rename" formaction="{% url 'plugins:slurpit_netbox:onboard' %}?remove=true&onboarded=true&return_url={% url 'plugins:slurpit_netbox:slurpitimporteddevice_list' %}{% if request.GET %}?{{ request.GET.urlencode }}{% endif %}" class="btn btn-danger ms-2 protect-btn" >
                <i class="mdi mdi-close" aria-hidden="true"></i> {% trans "Remove" %}
              </button>
          </div>
          {% endblock %}
      </div>
    </div>
    <p id="syncStatusCaption">
    </p>
     {# "Select all" form #}

      <div id="select-all-box" class="d-none card noprint">
        <div class="form col-md-12">
          <div class="card-body">
            <div class="float-end">
              {% if 'bulk_edit' in actions %}
                {% bulk_edit_button model query_params=request.GET %}
              {% endif %}
              {% if 'bulk_delete' in actions %}
                {% bulk_delete_button model query_params=request.GET %}
              {% endif %}
            </div>
            <div class="form-check">
              <input type="checkbox" id="select-all" name="_all" class="form-check-input" />
              <label for="select-all" class="form-check-label">
                {% blocktrans trimmed with count=table.rows|length object_type_plural=table.data.verbose_name_plural %}
                  Select <strong>all of the {{ count }} {{ object_type_plural }}</strong>
                {% endblocktrans %}
              </label>
            </div>
          </div>
        </div>
      </div>

     <div class="form form-horizontal">
       <div class="card">
         <div class="card-body htmx-container" id="object_list">
           {% include 'htmx/table.html' %}
         </div>
       </div>
     </div>
  </form>
 </div>