<ul class="list-inline mb-0 fs-3">
    <li class="list-inline-item">
      <a href="https://slurpit.io" target="_blank" class="link-secondary" rel="noopener">
        https://slurpit.io
      </a>
    </li>
    
  </ul>
  <ul class="list-inline list-inline-dots fs-4 mb-0" id="footer-stamp" hx-swap-oob="true">
    <li class="list-inline-item">
      {% now 'Y-m-d H:i:s T' %}
    </li>
    <li class="list-inline-item">
      {{ settings.HOSTNAME }} (v{{ settings.VERSION }})
    </li>
    <li class="list-inline-item">
      Slurpit(v{{ settings.PLUGINS_CONFIG.slurpit_netbox.version }})
    </li>
    
  </ul>